@@ -11,6 +11,7 @@
 #include <iostream>
 #include "../matrix/matrix.h"
 #include <limits>
+#include <array>
 
 #if defined(_MSC_VER) && _MSC_VER < 1400
 // Despite my efforts to disabuse visual studio of its usual nonsense I can't find a 
@@ -1274,6 +1275,61 @@ namespace dlib
     typedef vector<long,2> point;
     typedef vector<double,2> dpoint;
 
+// ----------------------------------------------------------------------------------------
+
+    inline bool is_convex_quadrilateral (
+        const std::array<dpoint,4>& pts
+    )
+    {
+        auto orientation = [&](size_t i)
+        {
+            size_t a = (i+1)%4;
+            size_t b = (i+3)%4;
+            return (pts[a]-pts[i]).cross(pts[b]-pts[i]).z();
+        };
+
+        // If pts has any infinite points then this isn't a valid quadrilateral.
+        for (auto& p : pts)
+        {
+            if (p.x() == std::numeric_limits<double>::infinity())
+                return false;
+            if (p.y() == std::numeric_limits<double>::infinity())
+                return false;
+        }
+
+        double s0 = orientation(0); 
+        double s1 = orientation(1); 
+        double s2 = orientation(2); 
+        double s3 = orientation(3); 
+
+        // if all these things have the same sign then it's convex.
+        return (s0>0&&s1>0&&s2>0&&s3>0) || (s0<0&&s1<0&&s2<0&&s3<0);
+    }
+
+// ----------------------------------------------------------------------------------------
+
+    template <
+        typename array_of_dpoints
+        >
+    inline double polygon_area (
+        const array_of_dpoints& pts
+    )
+    {
+        if (pts.size() <= 2)
+            return 0;
+
+        double val = 0;
+
+
+        for (size_t i = 1; i < pts.size(); ++i)
+            val += (double)pts[i].x()*pts[i-1].y() - pts[i].y()*pts[i-1].x();
+
+        const size_t end = pts.size()-1;
+        val += (double)pts[0].x()*pts[end].y() - pts[0].y()*pts[end].x();
+
+        return std::abs(val)/2.0;
+    }
+
 // ----------------------------------------------------------------------------------------
 
 }
