@@ -505,14 +505,30 @@ convergence:
         typename T,
         long NR,
         long NC,
-        typename MM,
-        typename L
+        typename MM
         >
     void orthogonalize (
-        matrix<T,NR,NC,MM,L>& m
+        matrix<T,NR,NC,MM,row_major_layout>& m
     )
     {
-        qr_decomposition<matrix<T,NR,NC,MM,L> >(m).get_q(m);
+        // We don't really need to use this temporary, but doing it this way runs a lot
+        // faster.
+        matrix<T,NR,NC,MM,column_major_layout> temp;
+        qr_decomposition<matrix<T,NR,NC,MM,row_major_layout>>(m).get_q(temp);
+        m = temp;
+    }
+
+    template <
+        typename T,
+        long NR,
+        long NC,
+        typename MM
+        >
+    void orthogonalize (
+        matrix<T,NR,NC,MM,column_major_layout>& m
+    )
+    {
+        qr_decomposition<matrix<T,NR,NC,MM,column_major_layout>>(m).get_q(m);
     }
 
 // ----------------------------------------------------------------------------------------
@@ -708,6 +724,74 @@ convergence:
 
 // ----------------------------------------------------------------------------------------
 
+    namespace simpl
+    {
+        template <
+            typename sparse_vector_type, 
+            typename T,
+            long Unr, long Unc,
+            long Wnr, long Wnc,
+            long Vnr, long Vnc,
+            typename MM,
+            typename L
+            >
+        void svd_fast (
+            bool compute_u,
+            const std::vector<sparse_vector_type>& A,
+            matrix<T,Unr,Unc,MM,L>& u,
+            matrix<T,Wnr,Wnc,MM,L>& w,
+            matrix<T,Vnr,Vnc,MM,L>& v,
+            unsigned long l,
+            unsigned long q 
+        )
+        {
+            const long n = max_index_plus_one(A);
+            const unsigned long k = std::min(l, std::min<unsigned long>(A.size(),n));
+
+            DLIB_ASSERT(l > 0 && A.size() > 0 && n > 0, 
+                "\t void svd_fast()"
+                << "\n\t Invalid inputs were given to this function."
+                << "\n\t l: " << l 
+                << "\n\t n (i.e. max_index_plus_one(A)): " << n 
+                << "\n\t A.size(): " << A.size() 
+                );
+
+            matrix<T,0,0,MM,L> Q;
+            find_matrix_range(A, k, Q, q);
+
+            // Compute trans(B) = trans(Q)*A.   The reason we store B transposed
+            // is so that when we take its SVD later using svd3() it doesn't consume
+            // a whole lot of RAM.  That is, we make sure the square matrix coming out
+            // of svd3() has size lxl rather than the potentially much larger nxn.
+            matrix<T,0,0,MM> B;
+            dlib::mutex mut;
+            parallel_for_blocked(0, A.size(), [&](long begin, long end)
+            {
+                matrix<T,0,0,MM> Blocal(n,k);
+                Blocal = 0;
+                for (long m = begin; m < end; ++m)
+                {
+                    for (unsigned long r = 0; r < k; ++r)
+                    {
+                        for (auto& i : A[m])
+                        {
+                            const auto c = i.first;
+                            const auto val = i.second;
+
+                            Blocal(c,r) += Q(m,r)*val;
+                        }
+                    }
+                }
+                auto_mutex lock(mut);
+                B += Blocal;
+            },1);
+
+            svd3(B, v,w,u);
+            if (compute_u)
+                u = Q*u;
+        }
+    }
+
     template <
         typename sparse_vector_type, 
         typename T,
@@ -726,49 +810,27 @@ convergence:
         unsigned long q = 1
     )
     {
-        const long n = max_index_plus_one(A);
-        const unsigned long k = std::min(l, std::min<unsigned long>(A.size(),n));
+        simpl::svd_fast(true, A,u,w,v,l,q);
+    }
 
-        DLIB_ASSERT(l > 0 && A.size() > 0 && n > 0, 
-            "\t void svd_fast()"
-            << "\n\t Invalid inputs were given to this function."
-            << "\n\t l: " << l 
-            << "\n\t n (i.e. max_index_plus_one(A)): " << n 
-            << "\n\t A.size(): " << A.size() 
-            );
-
-        matrix<T,0,0,MM,L> Q;
-        find_matrix_range(A, k, Q, q);
-
-        // Compute trans(B) = trans(Q)*A.   The reason we store B transposed
-        // is so that when we take its SVD later using svd3() it doesn't consume
-        // a whole lot of RAM.  That is, we make sure the square matrix coming out
-        // of svd3() has size lxl rather than the potentially much larger nxn.
-        matrix<T,0,0,MM> B;
-        dlib::mutex mut;
-        parallel_for_blocked(0, A.size(), [&](long begin, long end)
-        {
-            matrix<T,0,0,MM> Blocal(n,k);
-            Blocal = 0;
-            for (long m = begin; m < end; ++m)
-            {
-                for (unsigned long r = 0; r < k; ++r)
-                {
-                    for (auto& i : A[m])
-                    {
-                        const auto c = i.first;
-                        const auto val = i.second;
-
-                        Blocal(c,r) += Q(m,r)*val;
-                    }
-                }
-            }
-            auto_mutex lock(mut);
-            B += Blocal;
-        },1);
-
-        svd3(B, v,w,u);
-        u = Q*u;
+    template <
+        typename sparse_vector_type, 
+        typename T,
+        long Wnr, long Wnc,
+        long Vnr, long Vnc,
+        typename MM,
+        typename L
+        >
+    void svd_fast (
+        const std::vector<sparse_vector_type>& A,
+        matrix<T,Wnr,Wnc,MM,L>& w,
+        matrix<T,Vnr,Vnc,MM,L>& v,
+        unsigned long l,
+        unsigned long q = 1
+    )
+    {
+        matrix<T,0,0,MM,L> u;
+        simpl::svd_fast(false, A,u,w,v,l,q);
     }
 
 // ----------------------------------------------------------------------------------------
@@ -1141,68 +1203,68 @@ convergence:
 
         if (A.nr() > 4) // Only test for banded matrix if matrix is big enough
         {
-           // Detect if matrix is banded and, if so, matrix bandwidth
-           banded = true;
-           for (long r = 0; r < A.nr(); ++r)
-              for (long c = (r + bandwidth + 1); c < A.nc(); ++c)
-                 if (A(r, c) != 0)
-                 {
-                    bandwidth = c - r;
-                    if (bandwidth > A.nr() / 2)
+            // Detect if matrix is banded and, if so, matrix bandwidth
+            banded = true;
+            for (long r = 0; r < A.nr(); ++r)
+                for (long c = (r + bandwidth + 1); c < A.nc(); ++c)
+                    if (A(r, c) != 0)
                     {
-                       banded = false;
-                       goto escape_banded_detection;
+                        bandwidth = c - r;
+                        if (bandwidth > A.nr() / 2)
+                        {
+                            banded = false;
+                            goto escape_banded_detection;
+                        }
                     }
-                 }
         }
 escape_banded_detection:
 
         if (banded)
         {
-           // Store in compact form - use column major for LAPACK
-           matrix<T,0,0,default_memory_manager,column_major_layout> B(bandwidth + 1, A.nc());
-           set_all_elements(B, 0);
+            // Store in compact form - use column major for LAPACK
+            matrix<T,0,0,default_memory_manager,column_major_layout> B(bandwidth + 1, A.nc());
+            set_all_elements(B, 0);
 
-           for (long r = 0; r < A.nr(); ++r)
-              for (long c = r; c < std::min(r + bandwidth + 1, A.nc()); ++c)
-                 B(c - r, r) = A(r, c);
+            for (long r = 0; r < A.nr(); ++r)
+                for (long c = r; c < std::min(r + bandwidth + 1, A.nc()); ++c)
+                    B(c - r, r) = A(r, c);
 
 #ifdef DLIB_USE_LAPACK 
 
-           lapack::pbtrf('L', B);
-           
+            lapack::pbtrf('L', B);
+
 #else
 
-           // Peform compact Cholesky
-           for (long k = 0; k < A.nr(); ++k)
-           {
-              long last = std::min(k + bandwidth, A.nr() - 1) - k;
-              for (long j = 1; j <= last; ++j)
-              {
-                 long i = k + j;
-                 for (long c = 0; c <= (last - j); ++c)
-                    B(c, i) -= B(j, k) / B(0, k) * B(c + j, k);
-              }
-              T norm = std::sqrt(B(0, k));
-              for (long i = 0; i <= bandwidth; ++i)
-                 B(i, k) /= norm;
-           }
-           for (long c = A.nc() - bandwidth + 1; c < A.nc(); ++c)
-              B(bandwidth, c) = 0;
+            // Perform compact Cholesky
+            for (long k = 0; k < A.nr(); ++k)
+            {
+                long last = std::min(k + bandwidth, A.nr() - 1) - k;
+                for (long j = 1; j <= last; ++j)
+                {
+                    long i = k + j;
+                    for (long c = 0; c <= (last - j); ++c)
+                        B(c, i) -= B(j, k) / B(0, k) * B(c + j, k);
+                }
+                T norm = std::sqrt(B(0, k));
+                for (long i = 0; i <= bandwidth; ++i)
+                    B(i, k) /= norm;
+            }
+            for (long c = A.nc() - bandwidth + 1; c < A.nc(); ++c)
+                B(bandwidth, c) = 0;
 
 #endif
 
-           // Unpack lower triangular area
-           set_all_elements(L, 0);
-           for (long c = 0; c < A.nc(); ++c)
-              for (long i = 0; i <= bandwidth; ++i)
-              {
-                 long ind = c + i;
-                 if (ind < A.nc())
-                    L(ind, c) = B(i, c);
-              }
+            // Unpack lower triangular area
+            set_all_elements(L, 0);
+            for (long c = 0; c < A.nc(); ++c)
+                for (long i = 0; i <= bandwidth; ++i)
+                {
+                    long ind = c + i;
+                    if (ind < A.nc())
+                        L(ind, c) = B(i, c);
+                }
 
-           return L;
+            return L;
         }
 
 #ifdef DLIB_USE_LAPACK        
